# -*- coding: utf-8; -*-
################################################################################
#
#  WuttJamaican -- Base package for Wutta Framework
#  Copyright © 2023-2024 Lance Edgar
#
#  This file is part of Wutta Framework.
#
#  Wutta Framework is free software: you can redistribute it and/or modify it
#  under the terms of the GNU General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option) any
#  later version.
#
#  Wutta Framework is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
#  more details.
#
#  You should have received a copy of the GNU General Public License along with
#  Wutta Framework.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Batch Handlers
"""

from wuttjamaican.app import GenericHandler


class BatchHandler(GenericHandler):
    """
    Base class and *partial* default implementation for :term:`batch
    handlers <batch handler>`.

    This handler class "works as-is" but does not actually do
    anything.  Subclass must implement logic for various things as
    needed, e.g.:

    * :attr:`model_class`
    * :meth:`init_batch()`
    * :meth:`should_populate()`
    * :meth:`populate()`
    * :meth:`refresh_row()`
    """

    @property
    def model_class(self):
        """
        Reference to the batch :term:`data model` class which this
        batch handler is meant to work with.

        This is expected to be a subclass of
        :class:`~wuttjamaican.db.model.batch.BatchMixin` (among other
        classes).

        Subclass must define this; default is not implemented.
        """
        raise NotImplementedError("You must set the 'model_class' attribute "
                                  f"for class '{self.__class__.__name__}'")

    def make_batch(self, session, progress=None, **kwargs):
        """
        Make and return a new batch (:attr:`model_class`) instance.

        This will create the new batch, and auto-assign its
        :attr:`~wuttjamaican.db.model.batch.BatchMixin.id` value
        (unless caller specifies it) by calling
        :meth:`consume_batch_id()`.

        It then will call :meth:`init_batch()` to perform any custom
        initialization needed.

        Therefore callers should use this ``make_batch()`` method, but
        subclass should override :meth:`init_batch()` instead (if
        needed).

        :param session: Current :term:`db session`.

        :param progress: Optional progress indicator factory.

        :param \**kwargs: Additional kwargs to pass to the batch
           constructor.

        :returns: New batch; instance of :attr:`model_class`.
        """
        # generate new ID unless caller specifies
        if 'id' not in kwargs:
            kwargs['id'] = self.consume_batch_id(session)

        # make batch
        batch = self.model_class(**kwargs)
        self.init_batch(batch, session=session, progress=progress, **kwargs)
        return batch

    def consume_batch_id(self, session, as_str=False):
        """
        Fetch a new batch ID from the counter, and return it.

        This may be called automatically from :meth:`make_batch()`.

        :param session: Current :term:`db session`.

        :param as_str: Indicates the return value should be a string
           instead of integer.

        :returns: Batch ID as integer, or zero-padded 8-char string.
        """
        db = self.app.get_db_handler()
        batch_id = db.next_counter_value(session, 'batch_id')
        if as_str:
            return f'{batch_id:08d}'
        return batch_id

    def init_batch(self, batch, session=None, progress=None, **kwargs):
        """
        Initialize a new batch.

        This is called automatically from :meth:`make_batch()`.

        Default logic does nothing; subclass should override if needed.

        .. note::
           *Population* of the new batch should **not** happen here;
           see instead :meth:`populate()`.
        """

    def should_populate(self, batch):
        """
        Must return true or false, indicating whether the given batch
        should be populated from initial data source(s).

        So, true means fill the batch with data up front - by calling
        :meth:`do_populate()` - and false means the batch will start
        empty.

        Default logic here always return false; subclass should
        override if needed.
        """
        return False

    def do_populate(self, batch, progress=None):
        """
        Populate the batch from initial data source(s).

        This method is a convenience wrapper, which ultimately will
        call :meth:`populate()` for the implementation logic.

        Therefore callers should use this ``do_populate()`` method,
        but subclass should override :meth:`populate()` instead (if
        needed).

        See also :meth:`should_populate()` - you should check that
        before calling ``do_populate()``.
        """
        self.populate(batch, progress=progress)

    def populate(self, batch, progress=None):
        """
        Populate the batch from initial data source(s).

        It is assumed that the data source(s) to be used will be known
        by inspecting various properties of the batch itself.

        Subclass should override this method to provide the
        implementation logic.  It may populate some batches
        differently based on the batch attributes, or it may populate
        them all the same.  Whatever is needed.

        Callers should always use :meth:`do_populate()` instead of
        calling ``populate()`` directly.
        """

    def make_row(self, **kwargs):
        """
        Make a new row for the batch.  This will be an instance of
        :attr:`~wuttjamaican.db.model.batch.BatchMixin.__row_class__`.

        Note that the row will **not** be added to the batch; that
        should be done with :meth:`add_row()`.

        :returns: A new row object, which does *not* yet belong to any batch.
        """
        return self.model_class.__row_class__(**kwargs)

    def add_row(self, batch, row):
        """
        Add the given row to the given batch.

        This assumes a *new* row which does not yet belong to a batch,
        as returned by :meth:`make_row()`.

        It will add it to batch
        :attr:`~wuttjamaican.db.model.batch.BatchMixin.rows`, call
        :meth:`refresh_row()` for it, and update the
        :attr:`~wuttjamaican.db.model.batch.BatchMixin.row_count`.
        """
        session = self.app.get_session(batch)
        with session.no_autoflush:
            batch.rows.append(row)
            self.refresh_row(row)
        batch.row_count = (batch.row_count or 0) + 1

    def refresh_row(self, row):
        """
        Update the given batch row as needed, to reflect latest data.

        This method is a bit of a catch-all in that it could be used
        to do any of the following (etc.):

        * fetch latest "live" data for comparison with batch input data
        * calculate some data values based on the previous step
        * set row status based on other row attributes

        This method is called when the row is first added to the batch
        via :meth:`add_row()` - but may be called multiple times after
        that depending on the workflow.
        """
